using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.IO;
using System.Reflection;
using System.Threading;

using Pencil.Gaming;
using Pencil.Gaming.Audio;
using Pencil.Gaming.Graphics;
using Pencil.Gaming.MathUtils;
using Pencil.Gaming.Scripting;

namespace LudumDare26 {
	public class Program : Texture2D, IDisposable {
		public static Random RandomInstance { get; private set; }

		public const int WindowWidth = 800;
		public const int WindowHeight = 600;
		public const int Scale = 4;
		public const int PixelWidth = WindowWidth / Scale;
		public const int PixelHeight = WindowHeight / Scale;
		public const float PlayerRadius = 4f;
		public const float ClipDistance = 4f;

		private GameState state = GameState.Intro;

		public float PlayerZ { get; set; }
		public float PlayerX { get; set; }
		public Vector2 PlayerPosition {
			get { return new Vector2(PlayerX, PlayerZ); }
			set { 
				PlayerX = value.X;
				PlayerZ = value.Y;
			}
		}

		public float PlayerHealth { get; set; }
		private float redness = 0f;
		private float shotBrightness = 0f;

		//private Texture2D inventory;
		private Texture2D font;
		private Sound hurt;
		private Sound shoot;
		private int rectanglesKilled = 0;

		public Sound HealthPack { get; private set; }

		private int textureID = -1;

		private int mouseX = 0;
		public int MouseX {
			get { return mouseX; }
			set { mouseX = value; }
		}
		public float Rotation {
			get { return -mouseX / 500f; }
			//get { return -totalTime / 1f; }
			//get { return MathHelper.TauOver2; }
		}

		public float[] ZBuffer { get; private set; }
		public Enemy[] ForegroundBuffer { get; private set; }

		public List<Block> Blocks { get; private set; }
		public List<Enemy> Enemies { get; private set; }
		public List<HealthPack> HealthPacks { get; private set; }

		private Dictionary<char, Rectanglei>  fontChars = new Dictionary<char, Rectanglei>() {
			{ 'A', new Rectanglei(11 * 0, 14 * 0, 11, 14) },
			{ 'B', new Rectanglei(11 * 1, 14 * 0, 11, 14) },
			{ 'C', new Rectanglei(11 * 2, 14 * 0, 11, 14) },
			{ 'D', new Rectanglei(11 * 3, 14 * 0, 11, 14) },
			{ 'E', new Rectanglei(11 * 4, 14 * 0, 11, 14) },
			{ 'F', new Rectanglei(11 * 5, 14 * 0, 11, 14) },
			{ 'G', new Rectanglei(11 * 6, 14 * 0, 11, 14) },

			{ 'H', new Rectanglei(11 * 0, 14 * 1 + 1, 11, 14) },
			{ 'I', new Rectanglei(11 * 1, 14 * 1 + 1, 11, 14) },
			{ 'J', new Rectanglei(11 * 2, 14 * 1 + 1, 11, 14) },
			{ 'K', new Rectanglei(11 * 3, 14 * 1 + 1, 11, 14) },
			{ 'L', new Rectanglei(11 * 4, 14 * 1 + 1, 11, 14) },
			{ 'M', new Rectanglei(11 * 5, 14 * 1 + 1, 11, 14) },
			{ 'N', new Rectanglei(11 * 6, 14 * 1 + 1, 11, 14) },

			{ 'O', new Rectanglei(11 * 0, 14 * 2 + 1, 11, 14) },
			{ 'P', new Rectanglei(11 * 1, 14 * 2 + 1, 11, 14) },
			{ 'Q', new Rectanglei(11 * 2, 14 * 2 + 1, 11, 14) },
			{ 'R', new Rectanglei(11 * 3, 14 * 2 + 1, 11, 14) },
			{ 'S', new Rectanglei(11 * 4, 14 * 2 + 1, 11, 14) },
			{ 'T', new Rectanglei(11 * 5, 14 * 2 + 1, 11, 14) },
			{ 'U', new Rectanglei(11 * 6, 14 * 2 + 1, 11, 14) },

			{ 'V', new Rectanglei(11 * 0, 14 * 3 + 3, 11, 14) },
			{ 'W', new Rectanglei(11 * 1, 14 * 3 + 3, 11, 14) },
			{ 'X', new Rectanglei(11 * 2, 14 * 3 + 3, 11, 14) },
			{ 'Y', new Rectanglei(11 * 3, 14 * 3 + 3, 11, 14) },
			{ 'Z', new Rectanglei(11 * 4, 14 * 3 + 3, 11, 14) },

			{ '0', new Rectanglei(11 * 0, 14 * 4 + 4, 11, 14) },
			{ '1', new Rectanglei(11 * 1, 14 * 4 + 4, 11, 14) },
			{ '2', new Rectanglei(11 * 2, 14 * 4 + 4, 11, 14) },
			{ '3', new Rectanglei(11 * 3, 14 * 4 + 4, 11, 14) },
			{ '4', new Rectanglei(11 * 4, 14 * 4 + 4, 11, 14) },
			{ '5', new Rectanglei(11 * 5, 14 * 4 + 4, 11, 14) },
			{ '6', new Rectanglei(11 * 6, 14 * 4 + 4, 11, 14) },

			{ '7', new Rectanglei(11 * 0, 14 * 5 + 4, 11, 14) },
			{ '8', new Rectanglei(11 * 1, 14 * 5 + 4, 11, 14) },
			{ '9', new Rectanglei(11 * 2, 14 * 5 + 4, 11, 14) },
		};

		public Program() : base(PixelWidth, PixelHeight) {
		}

		private float totalTime;

		public void HurtPlayer() {
			hurt.Play();
			redness = 1f;
		}

		private MouseState previousMS = MouseState.GetMouseState();
		private MouseState currentMS;
		private void UpdateGun() {
			currentMS = MouseState.GetMouseState();
			if (currentMS.LeftButton && !previousMS.LeftButton) {
				shoot.Play();

				shotBrightness = .3f;

				int crossHairX = PixelWidth / 2;
				int crossHairY = PixelHeight / 2;

				Enemy enemy = GetForegroundBufferPixel(crossHairX, crossHairY);
				
				if (enemy != null) {
					++rectanglesKilled;

					
					restart:
					Enemy enem = new Enemy(this, RandomInstance.Next(-50, 50) + .5f, RandomInstance.Next(-50, 50) + .5f);
					foreach (Block b in Blocks) {
						if (b.X == -(int)enem.X - 1 && b.Z == (int)enem.Z) {
							goto restart;
						}
					}
					Enemies.Add(enem);

					if (RandomInstance.Next(4) == 0) {
						HealthPacks.Add(new HealthPack(this, enemy.X, enemy.Z));
					}
					Enemies.Remove(enemy);
				}
			}

			previousMS = currentMS;
		}

		private void UpdateMovement(float time) {
			Vector2 movement = Vector2.Zero;
			if (Glfw.GetKey('W')) {
				movement.Y += (float)Math.Cos(Rotation);
				movement.X -= (float)Math.Sin(Rotation);
			}
			if (Glfw.GetKey('S')) {
				movement.Y -= (float)Math.Cos(Rotation);
				movement.X += (float)Math.Sin(Rotation);
			}
			if (Glfw.GetKey('A')) {
				movement.X -= (float)Math.Cos(Rotation);
				movement.Y -= (float)Math.Sin(Rotation);
			}
			if (Glfw.GetKey('D')) {
				movement.X += (float)Math.Cos(Rotation);
				movement.Y += (float)Math.Sin(Rotation);
			}

			if (movement != Vector2.Zero) {
				movement.Normalize();
			}

			movement *= time;

			foreach (Block b in Blocks) {
				if (b.CollidesWith(movement * PlayerRadius)) {
					return;
				}
			}

			PlayerX += movement.X;
			PlayerZ += movement.Y;
		}

		private void Update(float time) {
			totalTime += time;
			switch (state) {
			case GameState.GamePlay:
				if (PlayerHealth <= 0f) {
					state = GameState.EndScreen;
					Clear(0xFF000000);
				}

				if (redness > 0f) {
					redness -= time;
				} else {
					redness = 0f;
				}

				if (shotBrightness > 0f) {
					shotBrightness -= time * 2f;
				} else {
					shotBrightness = 0f;
				}

				UpdateMovement(time);
				UpdateGun();

				foreach (Enemy e in Enemies) {
					e.Update(time);
				}

				foreach (HealthPack hp in HealthPacks) {
					hp.Update(time);
				}

				for (int i = 0; i < HealthPacks.Count; ++i) {
					if (HealthPacks [i].ShouldBeRemoved) {
						HealthPacks.RemoveAt(i--);
					}
				}

				int dummy;
				Glfw.GetMousePos(out mouseX, out dummy);

				break;
			}
		}

		private void DrawFloorAndCeiling() {
			Fill((x, y) => {
				float y1 = y - (PixelHeight / 2);

				float zFinal = (PixelHeight / 2f) / Math.Abs(y1);

				SetZBufferPixel(x, y, zFinal/* - PlayerZ*/);
				SetForegroundBufferPixel(x, y, null);

				return 0xFFFFFFFF;
			}
			);
		}

		private void DrawHearts() {
			int health = (int)(PlayerHealth * 10f);
			for (int i = -5; i < health - 5; ++i) {
				SetPixel(4 * i + 1 + PixelWidth / 2, PixelHeight - 3, 0xFFFF0000);
				SetPixel(4 * i + 1 + PixelWidth / 2, PixelHeight - 2, 0xFFFF0000);
				SetPixel(4 * i + 2 + PixelWidth / 2, PixelHeight - 3, 0xFFFF0000);
				SetPixel(4 * i + 2 + PixelWidth / 2, PixelHeight - 2, 0xFFFF0000);
			}
			for (int i = -5 + health; i < 5; ++i) {
				SetPixel(4 * i + 1 + PixelWidth / 2, PixelHeight - 3, 0xFF888888);
				SetPixel(4 * i + 1 + PixelWidth / 2, PixelHeight - 2, 0xFF888888);
				SetPixel(4 * i + 2 + PixelWidth / 2, PixelHeight - 3, 0xFF888888);
				SetPixel(4 * i + 2 + PixelWidth / 2, PixelHeight - 2, 0xFF888888);
			}
		}

		private void DrawCrossHair() {
			const uint color = 0xFFFFFFFF;

			SetPixel(PixelWidth / 2 - 1, PixelHeight / 2, color);
			SetZBufferPixel(PixelWidth / 2 - 1, PixelHeight / 2, 1f);

			SetPixel(PixelWidth / 2, PixelHeight / 2 - 1, color);
			SetZBufferPixel(PixelWidth / 2, PixelHeight / 2 - 1, 1f);

			SetPixel(PixelWidth / 2 + 1, PixelHeight / 2, color);
			SetZBufferPixel(PixelWidth / 2 + 1, PixelHeight / 2, 1f);

			SetPixel(PixelWidth / 2, PixelHeight / 2 + 1, color);
			SetZBufferPixel(PixelWidth / 2, PixelHeight / 2 + 1, 1f);
		}

		private void DrawString(string text, Vector2i pos) {
			int column = pos.X;
			foreach (char ch in text) {
				if (ch == ' ') {
					column += 11;
					continue;
				}

				Rectanglei charRect = fontChars [ch];

				font.DrawTo(this, new Vector2i(column, pos.Y), charRect);

				column += charRect.Width;
			}
		}

		private bool hasDrawnByAntonijn = false;
		private bool hasDrawnHighScores = false;
		private void DrawEverything() {
			switch (state) {
			case GameState.GamePlay:
				DrawFloorAndCeiling();
				//if (!Glfw.GetKey('Q')) 
				{
					DrawWalls();
				}
				DrawEnemies();
				DrawCrossHair();
				DrawHearts();
			/*inventory.DrawTo(this, new Vector2i(PixelWidth / 2 - inventory.Width / 2, PixelHeight - inventory.Height), (x, y) => {
				SetZBufferPixel(x, y, 1f); 
			}
			);*/

				foreach (HealthPack hp in HealthPacks) {
					hp.Draw();
				}
				PostProcess();
				break;
			case GameState.Intro:
				Clear(0xFF000000);

				if (totalTime < 2.5f) {
					if (!hasDrawnByAntonijn) {
						shoot.Play();
						hasDrawnByAntonijn = true;
					}

					DrawString("A GAME BY ANTONIJN", new Vector2i(0, 10 + 14 * 4));
				}

				if (totalTime > 3f) {
					DrawString("USE WASD TO MOVE", new Vector2i(10, 10 + 14 * 0));
					DrawString("AND THE MOUSE TO", new Vector2i(10, 10 + 14 * 1));
					DrawString("LOOK AROUND", new Vector2i(10, 10 + 14 * 2));

					DrawString("USE LEFT MOUSE", new Vector2i(10, 10 + 14 * 3));
					DrawString("BUTTON TO SHOOT", new Vector2i(10, 10 + 14 * 4));

					DrawString("KILL RECTANGLES", new Vector2i(10, 10 + 14 * 6));
					DrawString("PRESS SPACE", new Vector2i(10, 10 + 14 * 7));
					DrawString("TO CONTINUE", new Vector2i(10, 10 + 14 * 8));

					if (Glfw.GetKey(Key.Space)) {
						state = GameState.GamePlay;
					}
				}

				break;
			case GameState.EndScreen:
				if (!hasDrawnHighScores) {
					DrawString("YOU KILLED " + rectanglesKilled.ToString(), new Vector2i(10, 10 + 14 * 0));
					DrawString("RECTANGLES", new Vector2i(10, 10 + 14 * 1));

					int previous;
					if (rectanglesKilled >= (previous = int.Parse(File.ReadAllText("highscore")))) {
						DrawString("NEW HIGHSCORE", new Vector2i(10, 10 + 14 * 3));
						DrawString("PREVIOUS", new Vector2i(10, 10 + 14 * 5));
						DrawString("HIGHSCORE WAS", new Vector2i(10, 10 + 14 * 6));
						DrawString(previous.ToString(), new Vector2i(10, 10 + 14 * 7));

						DrawString("EXIT USING ESC", new Vector2i(10, 10 + 14 * 9));

						File.WriteAllText("highscore", rectanglesKilled.ToString());
					}
					hasDrawnHighScores = true;
				}
				break;
			}
		}

		private void DrawEnemies() {
			foreach (Enemy e in Enemies) {
				e.Draw();
			}
		}

		private void DrawWalls() {
			foreach (Block b in Blocks) {
				b.Draw();
			}
		}

		private void Draw() {
			DrawEverything();
			
			GL.BindTexture(TextureTarget.TextureRectangle, textureID);
			GL.TexImage2D(TextureTarget.TextureRectangle, 0, PixelInternalFormat.Rgba, PixelWidth, PixelHeight, 0, PixelFormat.Bgra, PixelType.UnsignedByte, Pixels);

			GL.Begin(BeginMode.Quads);
			{
				GL.TexCoord2(new Vector2(0f, 0f));
				GL.Vertex2(new Vector2(0f, 0f));

				GL.TexCoord2(new Vector2(PixelWidth, 0f));
				GL.Vertex2(new Vector2(WindowWidth, 0f));

				GL.TexCoord2(new Vector2(PixelWidth, PixelHeight));
				GL.Vertex2(new Vector2(WindowWidth, WindowHeight));

				GL.TexCoord2(new Vector2(0f, PixelHeight));
				GL.Vertex2(new Vector2(0f, WindowHeight));
			}
			GL.End();

			GL.BindTexture(TextureTarget.TextureRectangle, 0);

			Glfw.SwapBuffers();
		}

		private void DisposeResource(IDisposable res) {
			if (res != null) {
				res.Dispose();
			}
		}

		public void Dispose() {
			DisposeResource(hurt);
			DisposeResource(shoot);
			GL.DeleteTextures(1, ref textureID);
		}

		private static void Main(string[] args) {
			using (Program p = new Program()) {
				p.Run();
			}
		}

		private void Init() {
			Blocks = new List<Block>();
			Enemies = new List<Enemy>();
			HealthPacks = new List<HealthPack>();

			RandomInstance = new Random();
			for (int i = 0; i < 2000; ++i) {
				Block block = new Block(this, RandomInstance.Next(-50, 50), RandomInstance.Next(-50, 50));
				if (block.X == 0 && block.Z == 0) {
					goto endloop;
				}

				foreach (Block b in Blocks) {
					if (b.X == block.X && b.Z == block.Z) {
						goto endloop;
					}
				}
				Blocks.Add(block);
				continue;
				endloop:
				--i;
			}
			for (int x = -50; x <= 50; ++x) {
				Blocks.Add(new Block(this, x, -50));
				Blocks.Add(new Block(this, x, 50));
			}
			for (int y = -50; y <= 50; ++y) {
				Blocks.Add(new Block(this, -50, y));
				Blocks.Add(new Block(this, 50, y));
			}
			for (int i = 0; i < 1500; ++i) {
				Enemy enemy = new Enemy(this, RandomInstance.Next(-50, 50) + .5f, RandomInstance.Next(-50, 50) + .5f);
				foreach (Block b in Blocks) {
					if (b.X == -(int)enemy.X - 1 && b.Z == (int)enemy.Z) {
						goto endloop;
					}
				}
				Enemies.Add(enemy);
				continue;
				endloop:
				--i;
			}
			

			ZBuffer = new float[PixelWidth * PixelHeight];
			ForegroundBuffer = new Enemy[PixelWidth * PixelHeight];
			
			PlayerHealth = 1f;

			PlayerX = .5f;
			PlayerZ = .5f;

			hurt = new Sound(Assembly.GetEntryAssembly().GetManifestResourceStream("LudumDare26.Hurt.wav"), "wav");
			shoot = new Sound(Assembly.GetEntryAssembly().GetManifestResourceStream("LudumDare26.Shoot.wav"), "wav");
			HealthPack = new Sound(Assembly.GetEntryAssembly().GetManifestResourceStream("LudumDare26.HealthPack.wav"), "wav");
			font = new Texture2D(Assembly.GetEntryAssembly().GetManifestResourceStream("LudumDare26.Font.png"));

			GL.Viewport(0, 0, WindowWidth, WindowHeight);
			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			GL.Ortho(0.0, WindowWidth, WindowHeight, 0.0, 0.0, 1.0);
			GL.MatrixMode(MatrixMode.Modelview);
			GL.Enable(EnableCap.TextureRectangle);
			GL.GenTextures(1, out textureID);
			GL.BindTexture(TextureTarget.TextureRectangle, textureID);
			GL.TexParameter(TextureTarget.TextureRectangle, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Nearest);
			GL.TexParameter(TextureTarget.TextureRectangle, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Nearest);
			GL.BindTexture(TextureTarget.TextureRectangle, 0);
		}

		private Random rand = new Random();
		private void PostProcess() {
			Fill((x, y) => {
				//if (!Glfw.GetKey('P')) 
				{
					byte a, r, g, b;
					ColorFromUint(out a, out r, out g, out b, GetPixel(x, y));
					float zBufferPix = 1f / GetZBufferPixel(x, y);
					float @static = Math.Min((float)rand.NextDouble() + Math.Max(PlayerHealth, 0f) * .9f, 1);
					zBufferPix *= zBufferPix;
					zBufferPix *= @static;
					zBufferPix *= (1 + shotBrightness);
					zBufferPix = Math.Min(zBufferPix, 1f);
					/*if (Glfw.GetKey('I')) {
						return UintFromColor(a, (byte)(255 - r * zBufferPix), (byte)(255 - g * zBufferPix * (1f - redness)), (byte)(255 - b * zBufferPix * (1f - redness)));
					}*/
					return UintFromColor(a, (byte)(r * zBufferPix), (byte)(g * zBufferPix * (1f - redness)), (byte)(b * zBufferPix * (1f - redness)));
				}
				return GetPixel(x, y);
			}
			);
		}




		public void Run() {
			try {
				Glfw.Init();

				Glfw.OpenWindowHint(OpenWindowHint.NoResize, 1);
				Glfw.OpenWindow(WindowWidth, WindowHeight, 8, 8, 8, 8, 24, 0, WindowMode.Window);
				Glfw.SetWindowTitle("Dungeon of the Rectangles");
#if !DEBUG
				Glfw.Disable(GlfwEnableCap.MouseCursor);
#endif
				Glfw.SwapInterval(true);

				Init();

				Glfw.PollEvents();
				Glfw.SetTime(0.0);
				while (Glfw.GetWindowParam(WindowParam.Opened) == 1 && !Glfw.GetKey(Key.Escape)) {
					float deltaT = (float)Glfw.GetTime();
					Glfw.SetTime(0.0);

					Update(deltaT);
					Draw();

					Glfw.PollEvents();
				}
			} finally {
				Glfw.Terminate();
			}
		}

		public float GetZBufferPixel(int x, int y) {
			return ZBuffer [y * PixelWidth + x];
		}
		public void SetZBufferPixel(int x, int y, float value) {
			ZBuffer [y * PixelWidth + x] = value;
		}

		public Enemy GetForegroundBufferPixel(int x, int y) {
			return ForegroundBuffer [y * PixelWidth + x];
		}
		public void SetForegroundBufferPixel(int x, int y, Enemy value) {
			ForegroundBuffer [y * PixelWidth + x] = value;
		}
	}
}
