using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.IO;
using System.Drawing;

using Pencil.Gaming;
using Pencil.Gaming.Audio;
using Pencil.Gaming.Graphics;
using Pencil.Gaming.MathUtils;
using Pencil.Gaming.Scripting;

namespace LudumDare26 {
	public class Texture2D {
		public uint[] Pixels { get; private set; }
		public int Width { get; private set; }
		public int Height { get; private set; }

		public Texture2D(int width, int height) {
			Pixels = new uint[width * height];
			Width = width;
			Height = height;
		}

		public Texture2D(Stream str) {
			using (Bitmap bmp = new Bitmap(str)) {
				Width = bmp.Width;
				Height = bmp.Height;

				Pixels = new uint[Width * Height];

				for (int x = 0; x < bmp.Width; ++x) {
					for (int y = 0; y < bmp.Height; ++y) {
						unsafe {
							int color = bmp.GetPixel(x, y).ToArgb();
							SetPixel(x, y, *(uint *)&color);
						}
					}
				}
			}
		}

		public uint GetPixel(int x, int y) {
			return Pixels [y * Width + x];
		}
		public void SetPixel(int x, int y, uint value) {
			Pixels [y * Width + x] = value;
		}
		public static void ColorFromUint(out byte a, out byte r, out byte g, out byte b, uint color) {
			a = (byte)((color >> 24) & 0xFF);
			r = (byte)((color >> 16) & 0xFF);
			g = (byte)((color >> 8) & 0xFF);
			b = (byte)((color >> 0) & 0xFF);
		}
		public static uint UintFromColor(byte a, byte r, byte g, byte b) {
			return (uint)((a << 24) | (r << 16) | (g << 8) | b);
		}

		public void DrawTo(Texture2D tex, Vector2i pos) {
			DrawTo(tex, pos, (x, y) => { });
		}

		public void DrawTo(Texture2D tex, Vector2i pos, Action<int, int> act) {
			for (int x = 0; x < Width; ++x) {
				for (int y = 0; y < Height; ++y) {
					tex.SetPixel(x + pos.X, y + pos.Y, GetPixel(x, y));
					act(x + pos.X, y + pos.Y);
				}
			}
		}

		public void DrawTo(Texture2D tex, Vector2i pos, Rectanglei region) {
			for (int x = region.Left; x < region.Right; ++x) {
				for (int y = region.Top; y < region.Bottom; ++y) {
					tex.SetPixel(x + pos.X - region.Left, y + pos.Y - region.Top, GetPixel(x, y));
				}
			}
		}

		public void Clear(uint color) {
			Fill((x, y) => color);
		}

		public void Fill(Func<int, int, uint> colorFunc) {
			for (int x = 0; x < Width; ++x) {
				for (int y = 0; y < Height; ++y) {
					SetPixel(x, y, colorFunc(x, y));
				}
			}
		}
	}
}

